''' 
This module contains data structures for labelled Petri Nets.

Allows explicit place or transition ids for simpler comparison, especially
during testing.

For material on Petri Nets, see:
    - Quick and dirty introduction on wikipedia 
    https://en.wikipedia.org/wiki/Petri_net
   
Design Considerations:
    - The `Place` and `Transition` classes are hashable and comparable.
       - However, in order to support exporting and then re-importing a 
         Petri net, we use an identify attribute (pid/tid)
       - In python land, what you decide to use an identifier is up to you.
         - There are good cases for using an integer, say 4, or string, 
           say 'foo'.
         - These simple options make testing easier and repr's more readable.
       - However, if you plan to export and then reimport into ProM 
         or another tool (say pm4py), then you need to ensure that you 
         use a valid uuuid version 4 as the identifier. We offer support
         in various ways, one way is to leave the tid/pid as None and it 
         will be generated for you. 
            - The ProM implementation puts all elements into a hashmap
              based on the identifier, so it can silently break if you
              are not careful.
            - In many cases, you cannot even import without valid uuidv4s.
            - Similarly for pm4py, the identifier plays an important role
              if you plan to use their alignment tools.
        - Note, that hashing in python land occurs over the attributes of
          the node, so comparing a node with a identifier of an int 4 and
          a string "4" will be the same. We convert down to a string for
          identifiers to make life simple.
        - Also, hashing in python is salted so caching the hash is not 
          advised for threaded code. 
          See: https://docs.python.org/3.13/using/cmdline.html#cmdoption-R
'''
from collections.abc import Iterable
from copy import deepcopy
from typing import Union, FrozenSet, Dict, Set, Iterator
from uuid import uuid4

def verbosecmp(obj1:object,obj2:object) -> str:
    """
    This function produces a verbose statement about the given equality of
    the two objects.

    Returns a string statement about this equality.    
    """
    if obj1 == obj2:
        return "Same"
    result = ""
    if obj1.__dict__.keys() != obj2.__dict__.keys():
        ks1 = set(obj1.__dict__.keys())
        ks2 = set(obj2.__dict__.keys())
        result += f"Attributes not in both objects: {ks1 ^ ks2}\n"
    if obj1.__dict__ != obj2.__dict__:
        for k in obj1.__dict__:
            if k in obj2.__dict__ and obj1.__dict__[k] != obj2.__dict__[k] :
                result += f"Attribute {k} differs:\n"\
                        + f"  {obj1.__dict__[k]} vs\n  {obj2.__dict__[k]}\n"
    return result


class Place:
    """
    This a hashable and identifable place for a petri net.
    A place has a name and an identifier.
    """

    def __init__(self,name:str,pid:object=None):
        self._name = name
        # create an identifier.
        if (pid == None):
            self._pid = str(uuid4())
        else:
            self._pid = pid

    @property 
    def name(self) -> str:
        return self._name

    @property
    def pid(self) -> object:
        return self._pid

    @property
    def nodeId(self) -> str:
        return self._pid

    def __eq__(self,other) -> bool:
        if isinstance(other, type(self)):
            return self._name == other._name \
                and str(self._pid) == str(other._pid)
        return False

    def __hash__(self) -> int:
        return hash((self._name,str(self._pid)))

    def __str__(self) -> str:
        return f'({self.name})'  if self._pid is None \
               else f'({self.name}({self._pid}))'

    def __repr__(self) -> str:
        return f'Place("{self.name}",pid="{self.pid}")'


class Transition:
    """
    This is hashable and identifable transition for a Petri net.
    A transition has a name, an identifier, and can be silent.
    """

    def __init__(self,name:str,tid:str=None,silent:bool=False):
        self._name = name
        # create an identifier.
        if (tid == None):
            self._tid = str(uuid4())
        else:
            self._tid = tid
        # add extras
        self._silent = silent

    @property 
    def name(self) -> str:
        return self._name

    @property
    def silent(self) -> bool:
        return self._silent

    @property
    def tid(self) -> str:
        return self._tid

    @property
    def nodeId(self) -> str:
        return self._tid

    def __eq__(self,other) -> bool:
        if isinstance(other,type(self)):
            return self.name == other.name \
                and str(self.tid) == str(other.tid) \
                and self._silent == other._silent
        return False

    def __hash__(self) -> int:
        return hash((self._name,str(self._tid),self._silent))

    def __str__(self) -> str:
        return f'[{self.name}]'  if self._tid is None \
               else f'[{self.name}({self._tid})]'

    def __repr__(self) -> str:
        return f'Transition("{self.name}",tid="{self.tid}",' \
               + f'silent={self.silent})'

class Arc:
    """
    This is directed arc, which has no name or idenfitier, for Petri net.
    """

    def __init__(self,from_node:Union[Place,Transition],
                 to_node:Union[Place,Transition]):
        self._from_node = from_node
        self._to_node = to_node

    @property
    def from_node(self) -> Union[Place,Transition]:
        return self._from_node

    @property
    def to_node(self) -> Union[Place,Transition]:
        return self._to_node

    def __eq__(self,other) -> bool:
        if isinstance(other,type(self)):
            return self.from_node == other.from_node and \
                    self.to_node  == other.to_node
        return False

    def __hash__(self) -> int:
        return hash((self._from_node,self._to_node))

    def __str__(self) -> str:
        return f'{self.from_node} -> {self.to_node}' 

    def __repr__(self) -> str:
        return f'Arc(from_node={self.from_node.__repr__()},' \
               + f'to_node={self.to_node.__repr__()})'


class LabelledPetriNet:
    """
    A labelled Petri Net. It consists of places, transitions and directed arcs 
    between them.  Places and transitions have labels (names) and identifiers. 
    The net is also named.
    """

    def __init__(self, places:Iterable[Place], transitions:Iterable[Transition],
                 arcs:Iterable[Arc], 
                 name:str='Petri net'):
        self._places = set([ deepcopy(p) for p in places ])
        self._transitions = set([ deepcopy(t) for t in transitions])
        self._arcs = set([ deepcopy(a) for a in arcs])
        self._name = name
        self._postset_cache = dict(
            (node,self._postset(node)) 
            for node in 
            self._places.union(self._transitions)
        )
        self._preset_cache = dict(
            (node,self._preset(node))
            for node in
            self._places.union(self._transitions)
        )

    @property 
    def places(self) -> FrozenSet[Place]:
        return frozenset(self._places)

    @property
    def transitions(self) -> FrozenSet[Transition]:
        return frozenset(self._transitions)

    @property
    def arcs(self) -> FrozenSet[Arc]:
        return frozenset(self._arcs)

    @property
    def name(self) -> str:
        return self._name
    
    def postset(self,node:Union[Place,Transition]
        ) -> FrozenSet[Union[Place,Transition]]:
        """
        Returns the set of nodes that are reachable from the given node.
        """
        if node in self._postset_cache.keys():
            return self._postset_cache[node]
        else:
            raise ValueError(f"Node {node} does not belong to the net \n({self}).")
    
    def _postset(self,node:Union[Place,Transition]
        ) -> FrozenSet[Union[Place,Transition]]:
        """
        Computes the set of nodes that are reachable from the given node.
        """
        return frozenset([arc.to_node 
                          for arc in self.arcs 
                          if arc.from_node == node])
    
    def preset(self,node:Union[Place,Transition]
        ) -> FrozenSet[Union[Place,Transition]]:
        """
        Returns the set of nodes that can reach the given node.
        """
        if node in self._preset_cache.keys():
            return self._preset_cache[node]
        else:
            raise ValueError(f"Node {node} does not belong to the net \n({self}).")

    def _preset(self,node:Union[Place,Transition]
        ) -> FrozenSet[Union[Place,Transition]]:
        """
        Computes the set of nodes that can reach the given node.
        """
        return frozenset([arc.from_node 
                          for arc in self.arcs 
                          if arc.to_node == node])

    def __eq__(self,other) -> bool:
        if isinstance(other, type(self)):
            return self._name  == other._name and \
                   self._places == other._places and \
                   self._transitions == other._transitions and \
                   self._arcs   == self._arcs
        return False

    def __repr__(self) -> str:
        repr = "LabelledPetriNet(\n"
        # add places
        repr += "\tplaces=[\n"
        for p in self.places:
            repr += f"\t\t{p.__repr__()},\n"
        repr += "\t],\n"
        # add transitions
        repr += "\ttransitions=[\n"
        for t in self.transitions:
            repr += f"\t\t{t.__repr__()},\n"
        repr += "\t],\n"
        # add arcs
        repr += "\tarcs=[\n"
        for a in self.arcs:
            repr += f"\t\t{a.__repr__()},\n"
        repr += "\t],\n"
        # add name
        repr += f"\tname='{self.name}'\n"
        #close param
        repr += ")"
        return repr

    def __str__(self) -> str:
        _str = "LabelledPetriNet with name of '" + self._name + "'\n"
        _str += "\tPlaces: \n"
        for p in self.places:
            _str += f"\t\t- {p}\n"
        _str += "\tTransitions: \n"
        for t in self.transitions:
            _str += f"\t\t- {t}\n"
        _str += "\tArcs: \n"
        for a in self.arcs:
            _str += f"\t\t- {a}\n"
        return _str

class PetriNetMarking():
    """
    Data structure for a marking in a petri net, i.e. a multiset of places.
    """
    
    def __init__(self, marking:Dict[Place,int]) -> None:
        self._mark = deepcopy(marking)

    @property
    def mark(self) -> Dict[Place,int]:
        return deepcopy(self._mark)
    
    def contains(self, place:Place) -> bool:
        """
        Returns true if the place is in the marking.
        """
        return place in self
    
    # data-model functions
    def __getitem__(self, place:Place) -> int:
        if (self.contains(place)):
            return self._mark[place]
        return 0
    
    def __contains__(self, place:Place) -> None:
        if (isinstance(place, Place)):
            return place in self._mark and self._mark[place] > 0
        return False
    
    def __iter__(self) -> Iterator[Place]:
        return iter([p for p in self._mark if self._mark[p] > 0])
    
    ## self + other returns a new marking with the sum of the two
    def __add__(self, other:'PetriNetMarking') -> 'PetriNetMarking':
        if (not isinstance(other, type(self))):
            return NotImplemented
        new_mark = deepcopy(self._mark)
        for place in other._mark:
            if place in self:
                new_mark[place] = new_mark[place] + other._mark[place]
            else:
                new_mark[place] = other._mark[place]
        return PetriNetMarking(new_mark)
    
    ## self - other returns a new marking with the difference of the two
    def __sub__(self, other:'PetriNetMarking') -> 'PetriNetMarking':
        if (not isinstance(other, type(self))):
            return NotImplemented
        new_mark = deepcopy(self._mark)
        for place in other:
            if place in self:
                new_mark[place] = max(
                    new_mark[place] - other._mark[place],
                    0)
            else:
                new_mark[place] = 0
        return PetriNetMarking(new_mark)
    
    ## self << other checks that self is a subset of other
    def __lshift__(self, other:'PetriNetMarking') -> 'PetriNetMarking':
        if (not isinstance(other, type(self))):
            return NotImplemented
        for place in self:
            if place not in other._mark:
                return False
            if self._mark[place] > other._mark[place]:
                return False
        return True

    def __str__(self) -> str:
        vals = [ (i,v) for i,v in self._mark.items() if v > 0 ]
        return str(vals)
    
    def __eq__(self, other: object) -> bool:
        if (isinstance(other, type(self))):
            for place in self:
                if place not in other:
                    return False
                if self[place] != other[place]:
                    return False
            for place in other:
                if place not in self:
                    return False
            return True
        return False
    
    def __hash__(self) -> int:
        return hash(tuple(sorted(self._mark.items(),key=lambda x: x[0].name)))

class AcceptingPetriNet():
    """
    Abstraction to handle data for the execution of a `PetriNet`.
    """
    def __init__(self, net:LabelledPetriNet, inital:PetriNetMarking,
                 finals:Set[PetriNetMarking]) -> None:
        self._net = deepcopy(net)
        self._initial = deepcopy(inital)
        self._finals = deepcopy(finals)

    @property
    def net(self) -> LabelledPetriNet:
        return deepcopy(self._net)
    
    @property
    def initial_marking(self) -> PetriNetMarking:
        return deepcopy(self._initial)
    
    @property
    def final_markings(self) -> Set[PetriNetMarking]:
        return deepcopy(self._finals)
    
    def __eq__(self, other: object) -> bool:
        if (isinstance(other, type(self))):
            return self._net == other._net \
                and self._initial == other._initial \
                and self._finals == other._finals
    
class PetriNetSemantics():
    """
    Abstraction to handle the runtime execution semantics of a `PetriNet`.
    """
    def __init__(self, net:AcceptingPetriNet, curr:PetriNetMarking) -> None:
        self._anet = deepcopy(net)
        self._net = self._anet.net
        self._curr = deepcopy(curr)
        self._fires = self._firable()

    def fireable(self) -> Set[Transition]:
        """
        Returns the set of transitions that are fireable from the current marking.
        """
        return deepcopy(self._fires)

    def _firable(self) -> Set[Transition]:
        """
        Computes the set of transitions that are firable from the current marking.
        """
        ret = set()
        for tran in self._net.transitions:
            preset = self._net.preset(tran)
            preset = dict((f,1) for f in preset)
            if (PetriNetMarking(preset) << self._curr):
                ret.add(tran)
        return ret

    def peek(self, firing:Transition) -> 'PetriNetSemantics':
        """
        Peeks at the next marking that results from firing the given transition.
        """
        if (firing not in self._fires):
            raise ValueError("Given transition cannot fire from this marking.")
        postset = self._net.postset(firing)
        firing = dict( (f,1) for f in postset )
        firing = PetriNetMarking(firing)
        return PetriNetSemantics(self._net, self._curr - firing)

    def fire(self, firing:Transition) -> 'PetriNetSemantics':
        """
        Fires the given transition and returns the changed semantics.
        """
        if (firing not in self._fires):
            raise ValueError("Given transition cannot fire from this marking.")
        pre = dict( (f,1) for f in self._net.preset(firing) )
        mark = self._curr - PetriNetMarking(pre)
        post = dict( (f,1) for f in self._net.postset(firing) )
        mark = mark + PetriNetMarking(post)
        return PetriNetSemantics(self._anet, mark)

    def reached(self, mark:PetriNetMarking) -> bool:
        """
        Returns true if the given marking is reached.
        """
        return self._curr == mark

    # old interface
    def enabled(self) -> Set[Transition]:
        """
        returns the set of transitions that are enabled at this marking.
        """
        return self.fireable()

    def can_fire(self) -> Set[Transition]:
        """
        returns the set of transitions that can fire from this marking.
        """
        return self.enabled()
    
    def remark(self, firing:Transition) -> 'PetriNetSemantics':
        """
        Returns a new marking, that is one step from this marking by firing
        the given transition.
        """
        return self.fire(firing)
    
    def contains(self, place:Place) -> bool:
        """
        Returns true if the given place is in the marking.
        """
        if (place in self._mark.keys()):
            return self._mark[place] > 0
        return False
    
    def is_subset(self, other:'PetriNetMarking') -> bool:
        """
        Returns true if this marking is a subset of the other marking.
        """
        return self._mark << other
    
    def reached_final(self) -> bool:
        """
        Returns true if the marking is the final marking of the net.
        """
        for fmark in self._anet.final_markings:
            if (self.reached(fmark)):
                return True
        return False

    ## data-model functions
    def __str__(self):
        return f"{self._curr} -> {[ str(f) for f in self.fireable()]}"
    
    def __eq__(self, other: object) -> bool:
        if (isinstance(other, type(self))):
            return self._mark == other._mark and self._anet == other._anet
        return False
    
    def __hash__(self) -> int:
        return hash((self._mark,self._anet))
    
class ExecutablePetriNet():
    """
    Abstraction to contain all the data needed to execute a `PetriNet`.
    """
    def __init__(self, anet:AcceptingPetriNet, sem:PetriNetSemantics) -> None:
        self._anet = deepcopy(anet)
        self._sem = deepcopy(sem)

    @property
    def semantics(self) -> PetriNetSemantics:
        return deepcopy(self._sem)
    
    @property
    def anet(self) -> AcceptingPetriNet:
        return deepcopy(self._anet)
    
    ## data model functions
    def __eq__(self, other: object) -> bool:
        if (isinstance(other, type(self))):
            return self._anet == other._anet and self._sem == other._sem
        return False
    
    def __hash__(self) -> int:
        return hash((self._anet,self._sem))
    
def get_execution_semantics(anet:AcceptingPetriNet) -> ExecutablePetriNet:
    """
    Constructs an executable Petri net from the given Petri net and initial marking.
    """
    sem = PetriNetSemantics(anet, anet.initial_marking)
    return ExecutablePetriNet(anet, sem)



import os.path
import shutil
import subprocess
import tempfile
from logging import info
from typing import Dict
from warnings import warn

from pmkoalas.models.petrinets.pn import LabelledPetriNet,Place

DFENC='utf-8'

def dot_to_img(dot:str,dotf:str,imgf:str,outformat:str,dfenc:str=DFENC):
    with open(dotf,'w',encoding=dfenc) as outf:
        outf.write(dot)
        outf.write('\n')
    tfd, tfnamed = tempfile.mkstemp()
    tfi, tfnamei = tempfile.mkstemp()
    shutil.copy(dotf,tfnamed)
    cmdout = subprocess.run(['dot','-T' + outformat,'-o' + tfnamei, tfnamed])
    shutil.copy(tfnamei,imgf)

def export_DOT_to_image(vard:str,oname:str,dotStr:str,outformat:str='png'):
    dotf = os.path.join(vard,oname+'.dot')
    imgpnf = os.path.join(vard,oname+'_pn.' + outformat)
    info(f"Generating diagram {oname} ... ")
    dot_to_img(dotStr,dotf,imgpnf,outformat)


# colour scheme for lpns
START_PLACE_COLOUR = "#c5e1a5"
FINAL_PLACE_COLOUR = "#ef9a9a"
PLACE_COLOUR = "#ffe0b2"
TRANSITION_COLOUR = "#81c784"
TAU_COLUR = "lightgray"
ARC_COLOUR = "#263238"
BACKWARDS_COLOUR = "gray"
FONT_COLOUR = "black"

def prettier_handler_place(p:Place, lpn:LabelledPetriNet, nodeId) -> str:
    ret = ""
    label = '<'+'<BR/>'.join([p.name[i:i+16] for i in range(0, len(p.name), 16)])+">"
    if lpn.initial_marking is not None and lpn.initial_marking.contains(p):
        ret += f"\t\t{nodeId}[label={label},fillcolor=\"{START_PLACE_COLOUR}\"];"
    elif lpn.final_marking is not None and lpn.final_marking.contains(p):
        ret += f"\t\t{nodeId}[label={label},fillcolor=\"{FINAL_PLACE_COLOUR}\"];"
    else:
        ret += f"\t\t{nodeId}[label={label}];"
    ret += "\n"
    return ret

def lpn_prettier_dot(lpn:LabelledPetriNet):
    ret = "digraph{\n"
    ret += "\tdpi=150;rankdir=LR;nodesep=0.6;ranksep=0.3;\n"
    ret += "\tedge[penwidth=4,fontsize=16,minlen=2,fontname=\"roboto\"];\n"
    nodeIds = dict()
    nid = 0
    # here is where I would write a walking algorithm
    # but I am not ready to do that...
    if (False):
        # walks places (if marking is set)
        walked_places = []
        walked_transitions = []
        seen = set()
        ret += "\t{\n"
        ret += f"\t\tnode[shape=circle,margin=\"0.05, 0.05\",fillcolor=\"{PLACE_COLOUR}\",style=filled,width=1,penwidth=2,fontsize=9,fontname=\"roboto\"];\n" 
        # handle places
    else:
        ret += "\t{\n"
        ret += f"\t\tnode[shape=circle,margin=\"0.05, 0.05\",fillcolor=\"{PLACE_COLOUR}\",style=filled,width=1,penwidth=2,fontsize=9,fontname=\"roboto\"];\n"
        for v in lpn.places:
            nodeIds[v.nodeId] = f"n{nid}"
            ret += prettier_handler_place(v, lpn, nodeIds[v.nodeId])
            nid += 1
        assert nid == len(lpn.places)
        ret += "\t}\n"
        ret += "\t{\n"
        ret += f"\t\tnode[shape=box,style=\"rounded,filled\",fillcolor=\"{TRANSITION_COLOUR}\",width=2,penwidth=3,fontsize=18,fontname=\"roboto\",margin=\"0, 0.2\"];\n"
        for t in lpn.transitions:
            nodeIds[t.nodeId] = f"n{nid}"
            if (t.silent):
                ret += f"\t\t{nodeIds[t.nodeId]}[xlabel={t.name}, fillcolor=\"{TAU_COLUR}\", width=\"0.25\", height=1, margin=0, label=\"\"];\n" 
            else:
                label = '<'+'<BR/>'.join([t.name[i:i+12] for i in range(0, len(t.name), 12)])+">"
                ret += f"\t\t{nodeIds[t.nodeId]}[label={label}];\n"
            nid += 1
        assert nid == len(lpn.places) + len(lpn.transitions)
        ret += "\t}\n"
        elabel = 0
        for a in lpn.arcs:
            if a.from_node.nodeId in nodeIds:
                dotNode_from = nodeIds[a.from_node.nodeId]
            else:
                warn(f"Node {a.from_node} not found in nodeIds, likely something is wrong with the given LPN")
                dotNode_from = f"n{nid}"
                nid += 1
            if a.to_node.nodeId in nodeIds:
                dotNode_to = nodeIds[a.to_node.nodeId]
            else:
                warn(f"Node {a.to_node} not found in nodeIds, likely something is wrong with the given LPN")
                dotNode_to = f"n{nid}"
                nid += 1
            ret += f"\t{dotNode_from} -> {dotNode_to}[arrowhead=normal,color=\"{ARC_COLOUR}\"];\n"
            elabel += 1
        ret += "}"
    # assert elabel == len(lpn.arcs)
    return ret





"""
Module for storing meta classes and functions related to process discovery.
"""

from abc import ABC, abstractmethod

class DiscoveryTechnique(ABC):
    """
    Metaclass for process discovery techniques.

    This metaclass enforces the implementation of the `discover` method 
    in any subclass, which when called should invoke the discovery 
    procedure for the technique.
    """

    @abstractmethod
    def discover(self, log) -> object:
        """
        Abstract method for calling the procedure to discovery a graph or 
        model for the discovery technique.

        Subclasses must implement this method.
        """
        pass


"""
Module to control logging and logger instance so not to clash with other loggers
"""

import logging
from logging import Logger
from typing import Iterator
from functools import wraps
from sys import stdout

LOG_FRMT = '%(asctime)s|%(filename)-25s|%(funcName)-40s|%(levelname)-8s|: %(message)s'

fmt = LOG_FRMT
fmt_date = '%Y-%m-%dT%T'
formatter = logging.Formatter(fmt, fmt_date)
DEFAULT_HANDLER = logging.StreamHandler(stdout)
DEFAULT_HANDLER.setFormatter(formatter)

def get_logger() -> Logger:
    """
    This will get/create the unique logger for koalas.
    """

    # request or create a logger
    logger = logging.getLogger("koalas-log")

    # only do setup if needed
    if (not logger.hasHandlers()):
        logger.setLevel(logging.ERROR)
        logger.addHandler(DEFAULT_HANDLER)

    return logger 

class InfoIteratorProcessor():
    """
    A helper class to track the progress of an iterator.
    """
    LOGGER_MSG = "{name} :: {curr}/{size}"

    def __init__(self, itername:str, iter:Iterator, stack:int=3, size=None) -> None:
        self._iter = iter 
        self._name = itername 
        self._curr = 0
        self._size = len(iter) if size is None else size
        self._reportstack = stack

    def __iter__(self) -> Iterator:
        for step in self._iter:
            self._curr += 1
            yield step 
            self.update(stack_level=self._reportstack)

    def update(self, stack_level:int=2) -> None:
        get_logger().info(
            self.LOGGER_MSG.format(
                name = self._name,
                curr = self._curr,
                size = self._size
            )
            ,
            stacklevel=stack_level
        )


class InfoQueueProcessor():
    """
    A helper class to track the progress of queues.
    """
    LOGGER_MSG = "{name} :: {curr}/{size}"

    def __init__(self, itername:str, starting_size:int) -> None:
        self._processed = 0 
        self._name = itername
        self._size = starting_size

    def extent(self, extension:int) -> None:
        """
        Updates the length of the queue.
        """
        self._size += extension
        self.update(increase=0, stacklevel=3)

    def update(self, increase:int=1, stacklevel:int=2) -> None:
        """
        Updates the number of completed items from the queue and reports a
        message.
        """
        self._processed += increase

        get_logger().info(
            self.LOGGER_MSG.format(
                name = self._name,
                curr = self._processed,
                size = self._size
            )
            ,
            stacklevel=stacklevel
        )

def info(msg:str, *args, **kwargs):
    """
    Sends a info message to the koalas logger.
    """
    get_logger().info(msg, stacklevel=2, *args, **kwargs) 

def debug(msg:str, *args, **kwargs):
    """
    Sends a debug message to the koalas logger.
    """
    get_logger().debug(msg, stacklevel=2, *args, **kwargs) 

def warn(msg:str, *args, **kwargs):
    """
    Sends a warning message to the koalas logger.
    """
    get_logger().warn(msg, stacklevel=2, *args, **kwargs) 

def error(msg:str, *args, **kwargs):
    """
    Sends a error message to the koalas logger.
    """
    get_logger().error(msg, stacklevel=2, *args, **kwargs)

def setLevel(level):
    """
    Sets the logging level for koalas.
    """
    get_logger().setLevel(level)

def enable_logging(func):
    """
    Optional Parameters
    -------------------
    - debug [`True/False`]: if `True` function will show info level msgs
    - debug_level: sets the level of logging to show. 
    """
    # add to docstrings so 
    if (func.__doc__ != None):
        func.__doc__  += enable_logging.__doc__

    @wraps(func)
    def wrapped(*args, **kwargs):
        # keep the level before
        old_level = get_logger().getEffectiveLevel()
        # a tuple of the names of the parameters that func accepts
        func_params = func.__code__.co_varnames[:func.__code__.co_argcount]
        # grab all of the kwargs that are not accepted by func
        old_level = get_logger().level
        extra = set(kwargs.keys()) - set(func_params)
        debug = kwargs.pop("debug",False)
        level = kwargs.pop("debug_level", logging.INFO)
        # set logger settings
        if debug:
            # set debug on at info level
            get_logger().setLevel(level)
        # run function as intended
        val = func(*args, **kwargs)
        # reset logger
        get_logger().setLevel(old_level)
        # pass value back if needed
        return val
    
    return wrapped


"""
This module contains the abstraction layer that reads an arbitrary pnml file.
"""
from pmkoalas.models.petrinets.pn import LabelledPetriNet, Place
from pmkoalas.models.petrinets.pn import Transition, Arc
from pmkoalas.models.petrinets.pn import AcceptingPetriNet, PetriNetMarking
from pmkoalas.models.petrinets.dpn import PetriNetWithData, GuardedTransition
from pmkoalas.models.petrinets.dpn import AcceptingDataPetriNet
from pmkoalas.models.petrinets.wpn import WeightedPetriNet
from pmkoalas.models.petrinets.wpn import WeightedAcceptingPetriNet
from pmkoalas.models.petrinets.wpn import WeightedTransition
from pmkoalas.models.petrinets.guards import Guard,Expression

from os import path
from copy import deepcopy
from typing import Union
from xml.etree.ElementTree import parse

def parse_pnml_into_lpn(filepath:str,
        use_localnode_id:bool=False) -> Union[LabelledPetriNet,AcceptingPetriNet]:
    """
    Constructs a labelled Petri net from the given filepath to a pnml file.
    """
    # setup compontents
    initial_marking = {}
    final_marking = {}
    # begin parsing
    ## check that file exists
    if not path.exists(filepath):
        raise FileNotFoundError("pnml file not found at :: "+filepath)
    xml_tree = parse(filepath)
    pnml = xml_tree.getroot()
    net = pnml.find("net")
    net_name = net.find("name").find("text").text
    pages = net.findall("page")
    tgt_page = pages[0]
    ## we only parse the first page.
    places = {}
    place_ids = {}
    transitions = {}
    transition_ids = {}
    arcs = set()
    ### parse the places 
    for place in tgt_page.findall("place"):
        tools = place.findall("toolspecific")
        lid = None
        for tool in tools:
            if tool.attrib['tool'] == "ProM":
                if "localNodeID" in tool.attrib:
                    lid = tool.attrib["localNodeID"]
        id = place.attrib["id"]
        # making a place
        pid = lid if lid != None and use_localnode_id else id
        place_ids[id] = pid
        parsed = Place( place.find("name").find("text").text, pid)
        places[pid] = parsed
        # handle markings
        init = place.find("initialMarking")
        final = place.find("finalMarking")
        if init != None:
            initial_marking[places[pid]] = int(init.find("text").text)
        if final != None:
            final_marking[places[pid]] = int(final.find("text").text)
    ### parse the transitions
    for transition in tgt_page.findall("transition"):
        tools = transition.findall("toolspecific")
        lid = None
        for tool in tools:
            if tool.attrib['tool'] == "ProM":
                if "localNodeID" in tool.attrib:
                    lid = tool.attrib["localNodeID"]
        id = transition.attrib["id"]
        # check for silence 
        if "invisible"  in transition.attrib.keys():
            silent = transition.attrib["invisible"] == "true"
        else:
            silent = False
        # making a transition
        tid = lid if lid != None and use_localnode_id else id 
        transition_ids[id] = tid 
        parsed = Transition( 
            transition.find("name").find("text").text,
            tid,
            silent
        )
        transitions[tid] = parsed
    ### parse the arcs    
    nodes = deepcopy(places)
    nodes.update(transitions)
    node_ids = place_ids
    node_ids.update(transition_ids)
    for arc in tgt_page.findall("arc"):
        tools = arc.findall("toolspecific")
        lid = None
        for tool in tools:
            if tool.attrib['tool'] == "ProM":
                if "localNodeID" in tool.attrib:
                    lid = tool.attrib["localNodeID"]
        id = arc.attrib["id"]
        src = node_ids[arc.attrib["source"]]
        tgt = node_ids[arc.attrib["target"]]
        # making an arc 
        aid = lid if lid != None and use_localnode_id else id  # we aren't storing the actual id??
        src_node = nodes[src]
        tgt_node = nodes[tgt]
        arcs.add(Arc(
            src_node, tgt_node
        ))
    # finalise compontents
    net = LabelledPetriNet(
        places=set(list(places.values())),
        transitions=set(list(transitions.values())),
        arcs=set(list(arcs)),
        name=net_name
    )
    if (len(initial_marking.keys()) > 0) and (len(final_marking.keys()) > 0):
        net = AcceptingPetriNet(
            net,
            PetriNetMarking(initial_marking), 
            [PetriNetMarking(final_marking)]
        )
    return net

def parse_pnml_into_wpn(filepath:str,
                        use_localnode_id:bool=False
    ) -> Union[WeightedPetriNet,WeightedAcceptingPetriNet]:
    """
    Constructs a weighted Petri net from the given filepath to a pnml file.
    """
    # setup compontents
    initial_marking = {}
    final_marking = {}
    # begin parsing
    ## check that file exists
    if not path.exists(filepath):
        raise FileNotFoundError("pnml file not found at :: "+filepath)
    xml_tree = parse(filepath)
    pnml = xml_tree.getroot()
    net = pnml.find("net")
    net_name = net.find("name").find("text").text
    pages = net.findall("page")
    tgt_page = pages[0]
    ## we only parse the first page.
    places = {}
    place_ids = {}
    transitions = {}
    transition_ids = {}
    arcs = set()
    ### parse the places 
    for place in tgt_page.findall("place"):
        tools = place.findall("toolspecific")
        lid = None
        for tool in tools:
            if tool.attrib['tool'] == "ProM":
                if "localNodeID" in tool.attrib:
                    lid = tool.attrib["localNodeID"]
        id = place.attrib["id"]
        # making a place
        pid = lid if lid != None and use_localnode_id else id
        place_ids[id] = pid
        parsed = Place( place.find("name").find("text").text, pid)
        places[pid] = parsed
        # handle markings
        init = place.find("initialMarking")
        final = place.find("finalMarking")
        if init != None:
            initial_marking[places[pid]] = int(init.find("text").text)
        if final != None:
            final_marking[places[pid]] = int(final.find("text").text)
    ### parse the transitions
    for transition in tgt_page.findall("transition"):
        tools = transition.findall("toolspecific")
        lid = None
        weight = None
        for tool in tools:
            if tool.attrib['tool'] == "ProM":
                if "localNodeID" in tool.attrib:
                    lid = tool.attrib["localNodeID"]
            if tool.attrib['tool'] == "StochasticPetriNet":
                if "weight" in tool.attrib:
                    weight = float(tool.attrib["weight"])
        id = transition.attrib["id"]
        # check for silence 
        if "invisible"  in transition.attrib.keys():
            silent = transition.attrib["invisible"] == "true"
        else:
            silent = False
        # making a transition
        tid = lid if lid != None and use_localnode_id else id 
        transition_ids[id] = tid 
        parsed = WeightedTransition( 
            transition.find("name").find("text").text,
            tid,
            silent,
            weight if weight != None else 1
        )
        transitions[tid] = parsed
    ### parse the arcs    
    nodes = deepcopy(places)
    nodes.update(transitions)
    node_ids = place_ids
    node_ids.update(transition_ids)
    for arc in tgt_page.findall("arc"):
        tools = arc.findall("toolspecific")
        lid = None
        for tool in tools:
            if tool.attrib['tool'] == "ProM":
                if "localNodeID" in tool.attrib:
                    lid = tool.attrib["localNodeID"]
        id = arc.attrib["id"]
        src = node_ids[arc.attrib["source"]]
        tgt = node_ids[arc.attrib["target"]]
        # making an arc 
        aid = lid if lid != None and use_localnode_id else id  # we aren't storing the actual id??
        src_node = nodes[src]
        tgt_node = nodes[tgt]
        arcs.add(Arc(
            src_node, tgt_node
        ))
    # finalise compontents
    net = WeightedPetriNet(
        places=set(list(places.values())),
        transitions=set(list(transitions.values())),
        arcs=set(list(arcs)),
        name=net_name
    )
    if (len(initial_marking.keys()) > 0) and (len(final_marking.keys()) > 0):
        net = WeightedAcceptingPetriNet(
            net,
            PetriNetMarking(initial_marking), 
            [PetriNetMarking(final_marking)]
        )
    return net

def parse_pnml_for_dpn(filepath:str,
                       use_localnode_id:bool=False
    ) -> Union[PetriNetWithData, AcceptingDataPetriNet]:
    """
    Constructs a Petri net with data from the given filepath to a pnml file.
    """
    # setup compontents
    initial_marking = {}
    final_marking = {}
    # begin parsing
    ## check that file exists
    if not path.exists(filepath):
        raise FileNotFoundError("pnml file not found at :: "+filepath)
    xml_tree = parse(filepath)
    pnml = xml_tree.getroot()
    net = pnml.find("net")
    net_name = net.find("name").find("text").text
    pages = net.findall("page")
    tgt_page = pages[0]
    ## we only parse the first page.
    places = {}
    place_ids = {}
    transitions = {}
    transition_ids = {}
    arcs = set()
    ### parse the places 
    for place in tgt_page.findall("place"):
        tools = place.findall("toolspecific")
        lid = None
        for tool in tools:
            if tool.attrib['tool'] == "ProM":
                if "localNodeID" in tool.attrib:
                    lid = tool.attrib["localNodeID"]
        id = place.attrib["id"]
        # making a place
        pid = lid if lid != None and use_localnode_id else id
        place_ids[id] = pid
        parsed = Place( place.find("name").find("text").text, pid)
        places[pid] = parsed
        # handle markings
        init = place.find("initialMarking")
        final = place.find("finalMarking")
        if init != None:
            initial_marking[places[pid]] = int(init.find("text").text)
        if final != None:
            final_marking[places[pid]] = int(final.find("text").text)
    ### parse the transitions
    for transition in tgt_page.findall("transition"):
        tools = transition.findall("toolspecific")
        lid = None
        for tool in tools:
            if tool.attrib['tool'] == "ProM":
                if "localNodeID" in tool.attrib:
                    lid = tool.attrib["localNodeID"]
        id = transition.attrib["id"]
        # add guards if they exist
        if "guard" in transition.attrib:
            guard = Guard(Expression(transition.attrib['guard']))
        else: 
            guard = Guard(Expression("true"))
        # check for silence 
        if "invisible"  in transition.attrib.keys():
            silent = transition.attrib["invisible"] == "true"
        else:
            silent = False
        # making a transition
        tid = lid if lid != None and use_localnode_id else id
        transition_ids[id] = tid 
        parsed = GuardedTransition( 
            transition.find("name").find("text").text,
            guard,
            tid,
            silent
        )
        transitions[tid] = parsed
    ### parse the arcs    
    nodes = deepcopy(places)
    nodes.update(transitions)
    node_ids = place_ids
    node_ids.update(transition_ids)
    for arc in tgt_page.findall("arc"):
        tools = arc.findall("toolspecific")
        lid = None
        for tool in tools:
            if tool.attrib['tool'] == "ProM":
                if "localNodeID" in tool.attrib:
                    lid = tool.attrib["localNodeID"]
        id = arc.attrib["id"]
        src = node_ids[arc.attrib["source"]]
        tgt = node_ids[arc.attrib["target"]]
        # making an arc 
        aid = lid if lid != None and use_localnode_id else id
        src_node = nodes[src]
        tgt_node = nodes[tgt]
        arcs.add(Arc(
            src_node, tgt_node
        ))
    # finalise compontents
    dpn = PetriNetWithData(
        places=set(list(places.values())),
        transitions=set(list(transitions.values())),
        arcs=set(list(arcs)),
        name=net_name
    )
    if (len(initial_marking.keys()) > 0) and (len(final_marking.keys()) > 0):
        dpn = AcceptingDataPetriNet(
                dpn,
                PetriNetMarking(initial_marking), 
                [PetriNetMarking(final_marking)]
            )
    return dpn


"""
This module focuses on providing ways to produce a log from models in the class
of Petri nets.
"""
from copy import deepcopy
from typing import Dict, Iterable, Mapping,Set,List,Tuple

from pmkoalas import __version__
from pmkoalas.complex import ComplexEventLog, ComplexTrace, ComplexEvent
from pmkoalas.simple import Trace, EventLog
from pmkoalas.models.transitiontree import TransitionTreeGuard
from pmkoalas._logging import info, debug, InfoQueueProcessor, InfoIteratorProcessor
from pmkoalas.models.petrinets.pn import LabelledPetriNet, PetriNetMarking
from pmkoalas.models.petrinets.pn import AcceptingPetriNet, PetriNetSemantics
from pmkoalas.models.petrinets.pn import ExecutablePetriNet
from pmkoalas.models.petrinets.pn import Transition
from pmkoalas.models.petrinets.guards import Guard, GuardOutcomes
from pmkoalas.models.petrinets.dpn import AcceptingDataPetriNet
from pmkoalas.models.petrinets.dpn import get_execution_semantics
from pmkoalas._struct import Stack

class PlayoutTransitionGuard(TransitionTreeGuard):
    """
    Template for taking a guard off a transition in a petri net and converting
    it into the semantics for a transition tree.
    """

    def __init__(self, guard:'Guard', id:str=None) -> None:
        super().__init__()
        self._guard = guard
        self._id = 1 if id == None else id
        
    def check(self, data: ComplexEvent) -> 'GuardOutcomes':
        return self._guard.evaluate_data(data)
    
    def html_label(self) -> str:
        return f"g<sub>{self._id}</sub>"
    
    def expanded_html_label(self) -> str:
        return f"{self._guard}"
    
    def __hash__(self) -> int:
        return hash(self.expanded_html_label())

class PetriNetFiringSequence():
    """
    A data structure for a sequence of fired transitions.
    """

    def __init__(self, sem:PetriNetSemantics, fired:List['Transition'],
        final_marking:PetriNetMarking) -> None:
        self._final = final_marking
        self._mark = deepcopy(sem)
        self._seq = deepcopy(fired)

    def next(self) -> Set['Transition']:
        """
        Returns the set of transitions that can be seen as a possible next
        """
        # return nothing if we are at final
        if (self.reached_final()):
            return set()
        # remove previous linked silence from choices
        previous_silence = dict()
        for last in self._seq[-1::-1]:
            if last.silent:
                if last in previous_silence:
                    previous_silence[last] = previous_silence[last] + 1
                else:
                    previous_silence[last] = 1
            else:
                break
        drops = set()
        for silent in previous_silence:
            if previous_silence[silent] > 1:
                drops.add(silent)
        # present choices
        choices = self._mark.can_fire().difference(drops)
        return choices
    
    def fire(self, firing:'Transition') -> 'PetriNetFiringSequence':
        """
        Returns a new firing sequence, that is one step from this sequence
        by firing the given transition.
        """
        new_mark = self._mark.remark(firing)
        new_seq = self._seq + [firing]
        return PetriNetFiringSequence(
            new_mark, new_seq, self._final
        )
    
    def fired(self) -> List['Transition']:
        """
        Returns a sequence of transitions that have been fired. 
        """
        return deepcopy(self._seq)

    def reached_final(self) -> bool:
        """
        Returns true if the firing sequence reaches the final marking.
        """
        return self._mark.reached(self._final)
    
    def __len__(self) -> int:
        return len([ f for f in self._seq if not f.silent])
    
    def __hash__(self) -> int:
        return hash( tuple(self._seq + [self.reached_final()]))


class PlayoutEvent(ComplexEvent):
    """
    An abstraction for playout events.
    """

    def __init__(self, activity: str, guard:object) -> None:
        super().__init__(activity, {'guard' : guard})

    @property
    def guard(self) -> object:
        return self.data()['guard']
    
class PlayoutEnd(PlayoutEvent):
    """
    Dummy class to signal halting in playout.
    """

    def __init__(self) -> None:
        from pmkoalas.models.transitiontree import TransitionTreeGuard
        super().__init__("halt", TransitionTreeGuard())
    


class PlayoutTrace(ComplexTrace):
    """
    An abstraction for playout traces, with cuts and projections.
    """

    def __init__(self, 
                 events: Iterable[PlayoutEvent], 
                 data: Mapping[str, object] = None) -> None:
        super().__init__(events, data)

    def acut(self, i:int) -> 'Trace':
        """
        Returns the cut of activity sequence of this playout up to the
        i-th step (if it exists).
        """
        seq = self.simplify().sequence
        ret = []
        while i > 0 and len(seq) > 0:
            head = seq.pop(0)
            ret.append(head)
            i -= 1
        return Trace(ret)
    
    def act(self, i:int) -> str:
        """
        Returns the activity of the i-th step.
        """
        return self[i].activity()
    
    def guard(self, i:int) -> PlayoutTransitionGuard:
        """
        Returns the guard of the i-th step.
        """
        return self[i].guard
    
def generate_traces_from_lpn(
        model:ExecutablePetriNet, max_length:int,
        strict:bool=False, updates_on:int=1000) \
    -> EventLog:
    """
    Generates a simple log of a sample traces from a model with a length 
    of max_length. By default returns traces that did not reach the final
    marking. If strict is set to True, only traces that reach the final
    marking are returned.
    """
    inprogress:Stack[Tuple[Trace,PetriNetSemantics]] = Stack([ 
        (Trace([]), model.semantics) ]
    )
    completed = []
    pcount = 0
    while not inprogress.is_empty():
        top = inprogress.pop()
        trace:Trace = top[0]
        sem:PetriNetSemantics = top[1]
        debug(f"processing trace {str(trace)} with marking {str(sem)}")
        if len(trace) == max_length:
            if strict:
                if sem.reached_final():
                    completed.append(trace)
            else:
                completed.append(trace)
        else:
            if sem.reached_final():
                completed.append(trace)
            else:
                firable = sem.fireable()
                debug("firable transitions :: " + str(firable))
                for t in firable:
                    new_mark = sem.fire(t)
                    new_trace = Trace(trace.sequence + [t.name])
                    inprogress.push((new_trace, new_mark))
        pcount += 1
        if (pcount % updates_on == 0):
            debug(f"processing {len(inprogress)} partial executions of the lpn")
            debug(f"shortest trace :: {len(inprogress[0][0])}")
            debug(f"longest trace :: {len(inprogress[-1][0])}")
            pcount = 0
        info(f"generated traces :: {completed}")
    return EventLog(completed, f"Generated traces from LPN ({model.anet.net.name})")


def construct_playout_log(model:AcceptingDataPetriNet, max_length:int, 
        initial_marking:PetriNetMarking, final_marking:PetriNetMarking) \
        -> ComplexEventLog:
    """ 
    Constructs a log of play-out traces from a model with a length of 
    max_length + 1.
    """
    # importing here to remove circular dependencies.
    from pmkoalas.models.transitiontree import TransitionTreeMerge
    exec_model = get_execution_semantics(model)
    playout_traces = list()
    completed:Set[PetriNetFiringSequence] = set()
    incomplete = [ 
        PetriNetFiringSequence(exec_model.semantics, list(), final_marking)
    ]
    seen = set()
    pbar = InfoQueueProcessor(itername="processed partials",
                                 starting_size=len(incomplete)
    )
    while len(incomplete) > 0:
        select = incomplete.pop(0)
        if (len(select) > 0):
            completed.add(select)
        seen.add(select)
        potentials = list()
        for firing in select.next():
            potentials.append(select.fire(firing))
        for pot in potentials:
            if pot not in seen and len(pot) <= max_length:
                incomplete.append(pot)
                pbar.extent(1)
        pbar.update()        
    # now for each completed trace, we need to produce a complex trace
    trace_id = 1
    guard_id = 1
    guard_ids = {}
    for comp in InfoIteratorProcessor("play-out traces", completed):
        trace_map = {
            "concept:name" : f"play-out trace {trace_id}"
        }
        trace_seq = list()
        leftover_guard = None 
        # build sequence of fired transitions
        ## but profilerate guards on silence, without recording them
        for fired in comp.fired():
            if fired.guard not in guard_ids:
                guard_ids[fired.guard] = guard_id
                guard_id += 1
            if fired.silent:
                if leftover_guard == None:
                    leftover_guard = PlayoutTransitionGuard(
                        fired.guard,
                        guard_ids[fired.guard]
                    )
                else:
                    leftover_guard = TransitionTreeMerge(
                        leftover_guard, 
                        PlayoutTransitionGuard(fired.guard, guard_ids[fired.guard])
                    )
            else:
                guard = PlayoutTransitionGuard(fired.guard, guard_ids[fired.guard])
                if leftover_guard != None:
                    guard = TransitionTreeMerge(leftover_guard, guard)
                    leftover_guard = None
                trace_seq.append(
                    PlayoutEvent(fired.name, guard)
                )
        # add halt symbol if required
        if comp.reached_final():
            trace_seq.append(PlayoutEnd())
        # construct trace and store
        playout_traces.append(
            PlayoutTrace(
                trace_seq,
                trace_map
            )
        )
        trace_id += 1
    info("made playout log")
    return ComplexEventLog(playout_traces, dict({
        "meta:generated:by" : "pmkoalas",
        "meta:generator:version" : __version__
        }), 
        f"playout log for {model.net._name}"
    )


'''
This module contains typical useful restricts on collections in computer 
science.

Included are the following structures:
    - `Stack`
'''
from copy import deepcopy
from typing import Collection, Union

class Stack():
    '''
    A structure to represent a stack, where you can push, peek, or pop 
    elements on the stack.
    '''

    def __init__(self, starting:Collection) -> None:
        '''
        Initializes the stack with the starting collection, where the last
        element of the starting collection is at the top of the stack.
        '''
        self._stack = []
        for e in starting:
            self.push(e)

    def push(self, element) -> None:
        '''
        Pushes a copy of the element onto the stack.
        '''
        self._stack.append(deepcopy(element))

    def pop(self) -> Union[None, object]:
        '''
        Pop the top element off the stack, or return None when empty.
        '''
        if (len(self._stack) == 0):
            return None
        return self._stack.pop()
    
    def peek(self) -> Union[None, object]:
        '''
        Peek at the top element of the stack or returns None when empty.
        Returns a copy of the element.
        '''
        if (len(self._stack) == 0):
            return None
        return deepcopy(self._stack[-1])
    
    def is_empty(self) -> bool:
        '''
        Returns True if the stack is empty, False otherwise.
        '''
        return len(self._stack) == 0


"""
This module contains the abstractions and structures for a weighted Petri 
nets or stochastic Petri nets. These nets are used to model the likelihood
of actions in a systems at runtime.

See:
    - Process mining with labelled stochastic nets. Adam Burke. 2024.
    - Bause and Kritzinger (2002) - Stochastic Petri Nets: An Introduction 
    to the Theory. Freely available textbook at
    https://www.researchgate.net/publication/258705139_Stochastic_Petri_Nets_-An_Introduction_to_the_Theory
"""
from pmkoalas.models.petrinets.pn import Transition, Arc, Place
from pmkoalas.models.petrinets.pn import LabelledPetriNet, AcceptingPetriNet
from pmkoalas.models.petrinets.pn import PetriNetMarking, PetriNetSemantics
from pmkoalas.models.petrinets.pn import ExecutablePetriNet

from typing import Iterable, FrozenSet, Union, Set
class WeightedTransition(Transition):
    """
    A transition with an associated weight.
    """
    def __init__(self, name:str, tid:object = None, 
                 silent:bool = False, weight:float = 1):
        super().__init__(name, tid, silent)
        self._weight = weight

    @property
    def weight(self) -> float:
        return self._weight
    
    @weight.setter
    def weight(self,value:float) -> None:
        self._weight = value

    # data model methods
    def __str__(self) -> str:
        return f"[{self.name}({self._tid}) {self.weight}]"
    
    def __eq__(self,other) -> bool:
        if (super().__eq__(other)):
            return self.weight == other.weight
        return False

    def __hash__(self) -> int:
        return hash((super().__hash__(),self.weight))
    
    def __repr__(self) -> str:
        return f'WeightedTransition("{self.name}",tid="{self.tid}",' \
               + f'silent={self.silent}, weight={self.weight})'
    
class WeightedPetriNet(LabelledPetriNet):
    """
    A Petri net with transitions that have weights.
    """
    def __init__(self, places: Iterable[Place], 
                 transitions: Iterable[WeightedTransition], 
                 arcs: Iterable[Arc], 
                 name: str = 'Weighted Petri net'):
        super().__init__(places, transitions, arcs, name)
    
    @property
    def transitions(self) -> FrozenSet[WeightedTransition]:
        return frozenset([t for t in self._transitions])
    
    def preset(self, node:Union[Place,WeightedTransition]
               ) -> FrozenSet[Union[Place,WeightedTransition]]:
        return super().preset(node)
    
    def postset(self, node:Union[Place,WeightedTransition]
                ) -> FrozenSet[Union[Place,WeightedTransition]]:
        return super().postset(node)
    
    def __repr__(self) -> str:
        repr = "WeightedLabelledPetriNet(\n"
        # add places
        repr += "\tplaces=[\n"
        for p in self.places:
            repr += f"\t\t{p.__repr__()},\n"
        repr += "\t],\n"
        # add transitions
        repr += "\ttransitions=[\n"
        for t in self.transitions:
            repr += f"\t\t{t.__repr__()},\n"
        repr += "\t],\n"
        # add arcs
        repr += "\tarcs=[\n"
        for a in self.arcs:
            repr += f"\t\t{a.__repr__()},\n"
        repr += "\t],\n"
        # add name
        repr += f"\tname='{self.name}'\n"
        #close param
        repr += ")"
        return repr
    
    def __str__(self) -> str:
        _str = "WeightedLabelledPetriNet with name of '" + self._name + "'\n"
        _str += "\tPlaces: \n"
        for p in self.places:
            _str += f"\t\t- {p}\n"
        _str += "\tTransitions: \n"
        for t in self.transitions:
            _str += f"\t\t- {t}\n"
        _str += "\tArcs: \n"
        for a in self.arcs:
            _str += f"\t\t- {a}\n"
        return _str


class BuildablePetriNet(WeightedPetriNet):
    """
    Allows for the builder design pattern to be used for constructing a 
    Petri net. Users can add places, transitions and arcs through a single 
    chain of method calls. See usage below.

    Usage
    -----
    ```
    # setup elements
    buildable = BuildablePetriNet("dupe_tran_with_id")
    initial_place = Place("I",1)
    ta1 = Transition("a",1)
    ta2 = Transition("a",2)
    tb = Transition("b",3)
    finalPlace = Place("F",2)
    # build net
    buildable.add_place(initialPlace) \\
        .add_transition(ta1) \\
        .add_transition(ta2) \\
        .add_transition(tb) \\
        .add_place(finalPlace) \\
        .add_arc_between(initialPlace, ta1) \\
        .add_arc_between(ta1,finalPlace) \\
        .add_arc_between(initialPlace, ta2) \\
        .add_arc_between(ta2,finalPlace) \\
        .add_arc_between(initialPlace, tb) \\
        .add_arc_between(tb,finalPlace) 
    # get a state of build
    net = buildable.create_net()
    ```
    """
    def __init__(self,label:str=None):
        super().__init__(set(),set(),set(),label)

    def add_place(self,place:Place) -> 'BuildablePetriNet':
        "Adds a place to the net."
        self._places.add(place)
        return self

    def add_transition(self,tran:WeightedTransition) -> 'BuildablePetriNet':
        "Adds a tranistion to the net."
        self._transitions.add(tran)
        return self

    def add_arc(self,arc:Arc)-> 'BuildablePetriNet':
        "Adds an arc to the net."
        self._arcs.add(arc)
        return self

    def add_arc_between(self,from_node:Union[Place,WeightedTransition],
                             to_node:Union[Place,WeightedTransition]
                        ) -> 'BuildablePetriNet' :
        "Constructs an arc between the given nodes and adds it to the net."
        self.add_arc(Arc(from_node,to_node))
        return self

    def create_net(self) -> WeightedPetriNet:
        "Returns a Petri net of the current built state"
        return WeightedPetriNet(
            self._places,
            self._transitions,
            self._arcs,
            self._name
        )

    def __eq__(self,other):
        if isinstance(other,type(self)) or isinstance(other,LabelledPetriNet):
            return self._name  == other._name and \
                   self._places == other._places and \
                   self._transitions == other._transitions and \
                   self._arcs   == self._arcs
        return False

class WeightedAcceptingPetriNet(AcceptingPetriNet):
    """
    A weighted Petri net with marking information.
    """
    def __init__(self, net:WeightedPetriNet,
                 inital:PetriNetMarking, 
                 finals:Set[PetriNetMarking]):
        super().__init__(net, inital, finals)

class WeightedPetriNetSemantics(PetriNetSemantics):
    """
    The semantics of a weighted Petri net.
    """
    def __init__(self, wanet:WeightedAcceptingPetriNet, curr:PetriNetMarking):
        super().__init__(wanet, curr)

    def fireable(self) -> set[WeightedTransition]:
        return super().fireable()
    
    def peek(self, tran:WeightedTransition) -> 'WeightedPetriNetSemantics':
        ret = super().peek(tran)
        return WeightedPetriNetSemantics(ret._anet, ret._curr)
    
    def fire(self, tran:WeightedTransition) -> 'WeightedPetriNetSemantics':
        ret = super().fire(tran)
        return WeightedPetriNetSemantics(ret._anet, ret._curr)
    
class ExecutableWeightedPetriNet(ExecutablePetriNet):
    """
    An executable weighted Petri net.
    """
    def __init__(self, wanet:WeightedAcceptingPetriNet, 
                 sem:WeightedPetriNetSemantics):
        super().__init__(wanet, sem)

    @property
    def semantics(self) -> WeightedPetriNetSemantics:
        return self._sem
    
    @property
    def anet(self) -> WeightedAcceptingPetriNet:
        return self._anet

def get_execution_semantics(wanet:WeightedAcceptingPetriNet):
    """
    Returns the execution semantics of a given weighted Petri net.
    """
    sem = WeightedPetriNetSemantics(wanet, wanet.initial_marking)
    return ExecutableWeightedPetriNet(wanet, sem)


"""
This module contains an alignment like process for transition trees, called 
matchings.
"""
from typing import Any, Union, Dict, List, Set, Iterable, Tuple
from copy import deepcopy

from joblib import Parallel, delayed

from pmkoalas.models.transitiontree import TransitionTreeFlow
from pmkoalas.models.transitiontree import TransitionTree
from pmkoalas.models.transitiontree import TransitionTreeVertex
from pmkoalas.simple import Trace, EventLog
from pmkoalas.complex import ComplexEventLog
from pmkoalas._logging import info, InfoIteratorProcessor,InfoQueueProcessor
from pmkoalas._logging import debug

class Skipper():
    """
    A dummy class for the skip symbol in matchings, i.e. ">>"
    """

    def __str__(self) -> str:
        return ">>"
    
    def __repr__(self) -> str:
        return "Skipper()"
    
    def __hash__(self) -> int:
        return hash(str(self))
    
    def __eq__(self, other: object) -> bool:
        return type(self) == type(other)

class Path():
    """
    A sequence of interconnected flows and/or skips within a tree. 
    """

    def __init__(self, sequence:List[Union[Skipper,'TransitionTreeFlow']]) \
        -> None:
        self._seq = deepcopy(sequence)

    @property
    def sequence(self) -> List[Union[Skipper,'TransitionTreeFlow']]:
        return deepcopy(self._seq)
    
    @property
    def noskips(self) -> List['TransitionTreeFlow']:
        ret = []
        for p in self._seq:
            if not isinstance(p, Skipper):
                ret.append(deepcopy(p))
        return ret
    
    def __iter__(self) -> Iterable[Union[Skipper, TransitionTreeFlow]]:
        for s in self._seq:
            yield s
    
    def __str__(self) -> str:
        ret = "< "
        for s in self.sequence:
            add = ""
            if (isinstance(s,TransitionTreeFlow)):
                add = s.activity()
            else:
                add = str(s)
            ret += add + ","
        return ret[:-1] + " >"
    
    def __hash__(self) -> int:
        return hash(tuple(self.sequence))
    
    def __eq__(self, other: object) -> bool:
        if (type(self) == type(other)):
            return self.__hash__() == other.__hash__()
        return False
    
    def __len__(self) -> int:
        return len(self._seq)

class Matching():
    """
    A mapping data structure, which maps variants to a path in a tree.
    """

    def __init__(self, mapping:Dict[Trace,Path]):
        self._map = deepcopy(mapping) 

    def add_to_map(self, trace:Trace, path:Path) -> None:
        """
        Either adds a new variant to the mapping or swaps an existing mapping
        with the given trace and path.
        """
        self._map[trace] = path

    def __getitem__(self, trace:Trace) -> Union[Path,None]:
        if trace in self._map:
            return deepcopy(self._map[trace])
        else:
            return None
        
class ManyMatching(Matching):
    """
    A mapping data structure, which maps traces to a set of paths in a tree.
    """

    def __init__(self, mapping:Dict[Trace,Set[Path]]):
        super().__init__(mapping)

    def add_to_map(self, trace:Trace, paths:Union[Path,Set[Path]]) -> None:
        """
        Either introduces a new trace into the mapping or updates the existing
        mapping to a trace with given path/s.
        """
        if (trace not in self._map):
            self._map[trace] = set()
            if isinstance(paths, Path):
                self._map[trace].add(paths)
            else:
                self._map[trace] = paths 
        else:
            old_set = self._map[trace]
            if isinstance(paths, Path):
                old_set.add(paths)
            else:
                old_set = old_set.union(paths)

    def __getitem__(self, trace: Trace) -> Union[Set[Path], None]:
        return super().__getitem__(trace)
    
def find_path_in_tree(tree:TransitionTree, tgt:TransitionTreeVertex)\
    -> Path:
    """
    Constructs a path from the root to the given node in a tree.
    """
    seq = []
    breakdown = Trace([])
    goal = tgt.sigma_sequence().sequence
    while breakdown != tgt.sigma_sequence():
        step = Trace(breakdown.sequence + [goal.pop(0)])
        found = False
        for flow in tree.flows():
            if flow.offering().sigma_sequence() == breakdown:
                if flow.next().sigma_sequence() == step:
                    seq.append(flow)
                    breakdown = step 
                    found = True 
                    break 
        if (not found):
            raise ValueError("Unable to find a interconnected "\
                             +"sequence of flows.")
    return Path(seq)

def mutate_path_with_skips(path:Path, k:int) -> Set[Path]:
    """
    Constructs all mutations of the path with additional skips of at most
    length k.
    """
    
    if (len(path) > k):
        return set()
    elif len(path) == k:
        return set([path])
    else:
        ret = set()
        adding = set([path])
        log_que = InfoQueueProcessor("mutation size",1)
        while len(adding) > 0:
            mutaters = adding
            ret = ret.union(adding)
            adding = set()
            for muter in mutaters:
                info(f"mutating :: {muter}")
                for i in range(len(muter)+1):
                    spath = muter.sequence[:i] + [Skipper()] + muter.sequence[i:]
                    spath = Path(spath)
                    if (len(spath) <= k) and spath not in ret:
                        adding.add(spath)
                        info(f"adding new path to mutations :: {spath}")
                        log_que.extent(1)
                log_que.update(1)
        return ret

def find_all_paths(tree:TransitionTree, k:int) -> Set[Path]:
    """
    Finds all paths through or in the tree, for a given length.
    """
    seen_paths = set()
    # construct all paths to nodes of at most length k
    nodes = [ v for v in tree.vertices() if len(v.sigma_sequence()) <= k]
    complete_paths = [ find_path_in_tree(tree,n) for n in nodes]
    seen_paths = seen_paths.union(set(complete_paths))
    # constuct all paths with skips 
    for path in InfoIteratorProcessor("computing all mutations", 
                                      [ path for path in complete_paths ]):
        mutate = mutate_path_with_skips(path, k)
        seen_paths = seen_paths.union(mutate)
    return seen_paths

def cost_of_path(path:Path, trace:Trace, root_is_terminal:bool=False) -> int:
    """
    Computes of the cost of the path in the context of the given trace.
    An optimal path for a trace has a cost of zero, while a non-optimal
    path has a cost greater than zero.
    """
    len_cost = abs(len(trace) - len(path.noskips))
    act_cost = 0
    for act, actf in zip(trace, path):
        if isinstance(actf, Skipper):
            act_cost += 1
        else:
            act_cost += 0 if act == actf.activity() else 1
    if (len(path.noskips) > 0 ):
        term_cost = 0 if path.noskips[-1].next().terminal() else 1
    else:
        term_cost = 0 if root_is_terminal else 1
    return len_cost + act_cost + term_cost

def find_least_costy_paths(paths:Set[Path], trace:Trace, 
        root_is_terminal:bool=False) -> set[Path]:
    """
    Finds the subset of the given paths that are least costy.
    """
    ret = set()
    costing = dict( 
        (path, cost_of_path(path, trace, root_is_terminal=root_is_terminal))
        for path 
        in paths
    )
    ideal = -1
    while len(ret) == 0:
        ideal += 1
        ret = set()
        for path,cost in costing.items():
            if cost == ideal:
                ret.add(path) 
    return ret 
    
class ExpontentialPathWeighter():
    """
    A weighting function, that weights candidates with exponential offset for,
    non-optimal paths, between a path and a trace.
    """
    # offset cost
    kappa = 0.9

    def __init__(self, candidates:Set[Path]) -> None:
        self._cands = deepcopy(candidates)
        self._size = len(self._cands)

    def __call__(self, path:Path, trace:Trace) -> float:
        debug(f"computing weight for path {path} and trace {trace}")
        if path in self._cands:
            cost = cost_of_path(path, trace)
            debug(f"cost of path was {cost}")
            share = 1 / self.size
            offset = self.kappa ** cost
            debug(f"offset was {offset}, return {share * offset}")
            return share * offset
        else:
            return 0
    
    @property
    def size(self) -> int:
        return self._size
    
    @property
    def share(self) -> float:
        """
        The maximum share that a path can be given
        """
        return 1 / self.size
        
def _computation_many_matching(trace, tree, allcads) -> Tuple[Trace,Set[Path]]:
    """
    The individual work for a given trace, to find a matching.
    """
    # noskips = find_non_skipping_candidates(tree, trace)
    # skippings = find_skipping_candidates(tree,trace)
    # terminals = find_terminal_candidates(tree,trace)
    if len(allcads) > 0:
        least_costy = find_least_costy_paths(
            allcads,
            trace,
            root_is_terminal=tree.root() in tree.terminals()
        )
    else: 
        raise Exception(f"Unable to find any candidates for :: {trace}")
    cost = 0
    for path in least_costy:
        cost = cost_of_path(path, trace)
        break
    info(f"for trace {trace}, we found {len(least_costy)} paths of cost={cost}")
    return (trace, least_costy)

def construct_many_matching(log:Union[EventLog,ComplexEventLog], 
                            tree:TransitionTree) \
     -> ManyMatching:
    """
    Constructs a set of likely cadidate paths for each trace in the log,
    then constructs a map from traces to these sets.
    """
    mapping = ManyMatching(dict())
    max_length = max(
        [len(trace) for trace,_ in log]
    )
    allcads = find_all_paths(tree, max_length+1)
    rets = Parallel(n_jobs=-2)(
        delayed(_computation_many_matching)(trace,tree,allcads)
        for trace,allcads
        in InfoIteratorProcessor(
            "matchings", 
            [ (trace,set([ cad for cad in allcads if len(cad) <= len(trace)]))
             for trace,_ in log ]
            ,stack=8)
    )
    for trace,least_costy in rets:
        info(f"no. of matching generated for {trace} was {len(least_costy)}")
        mapping.add_to_map(trace, least_costy)
    return mapping




"""
This module includes some utility functions for Petri nets.
"""
from pmkoalas.models.petrinets.wpn import WeightedTransition

SILENT_TRANSITION_DEFAULT_NAME='tau'
def silent_transition(name=None,tid=None,weight=None):
    tn = SILENT_TRANSITION_DEFAULT_NAME
    if name:
        tn = name
    tw = 1
    ttid = None
    if tid:
        ttid = tid
    if weight:
        tw = weight
    return WeightedTransition(name=tn,weight=tw,tid=ttid,silent=True)

# Candidate to move to a utility package
def steq(self,other):
    if type(other) is type(self):
        return self.__dict__ == other.__dict__
    return False